from __future__ import print_function

import argparse
import os
import sys

from .filters import criticmarkup_accept_filter, criticmarkup_reject_filter, criticmarkup_tex_diff_filter, criticmarkup_html_diff_filter, markdown_filter, pandoc_filter, html_filter

from .version import __version__


def normalize_format(ext):
    if ext == 'md':
        return 'markdown'
    elif ext == 'tex':
        return 'latex'
    else:
        return ext


def main(args, body, output_format, is_binary):
    # parse CritiMarkup
    # diff/markup mode
    if args.critic_mode[0] in ('d', 'm'):
        if args.to in ('latex', 'pdf'):
            body = criticmarkup_tex_diff_filter(body)
        # for any other format, use HTML (many formats support inline HTML)
        else:
            body = criticmarkup_html_diff_filter(body)
    # accept mode
    elif args.critic_mode[0] == 'a':
        body = criticmarkup_accept_filter(body)
    # reject mode
    elif args.critic_mode[0] == 'r':
        body = criticmarkup_reject_filter(body)
    else:
        print('Unknown critic mode {}.'.format(args.critic_mode), file=sys.stderr)

    # convert between to and from format
    # only convert markdown to html or tex if the output extension is really that format
    if args.from_format == 'markdown' and output_format == 'html' and args.engine in ('markdown', 'markdown2'):
        body = markdown_filter(body, args.engine)
        body = html_filter(body, args.critic_template, args.critic_mode[0], args.standalone)
    elif output_format != args.from_format:
        # as long as the final format will be latex, don't add html_filter
        if args.to not in ('latex', 'pdf'):
            # defer standalone to pandoc
            body = html_filter(body, args.critic_template, args.critic_mode[0], False)
        if is_binary:
            # only pypandoc handles binary output
            body = pandoc_filter(body, args.from_format, output_format, args.standalone, 'pypandoc', outputfile=args.output)
        else:
            body = pandoc_filter(body, args.from_format, output_format, args.standalone, args.engine)
    elif args.to != 'latex':
        body = html_filter(body, args.critic_template, args.critic_mode[0], args.standalone)

    # write (if is binary, already written above)
    if not is_binary:
        args.output.write(body)


def get_args():
    parser = argparse.ArgumentParser(description='Convert Critic Markup.')

    parser.add_argument('input', type=argparse.FileType('r'), default=sys.stdin,
                        help='Input file. Default: stdin.')
    parser.add_argument('-o', '--output',
                        help='Output file. Default: stdout.')

    parser.add_argument('-t', '--to',
                        help='Output format. Default: inferred from --output.')
    parser.add_argument('-f', '--from', dest='from_format',
                        help='Input format. Default: inferred from --input.')
    parser.add_argument('-s', '--standalone', action='store_true',
                        help='Output standalone html.')

    parser.add_argument('-i', '--inplace', action='store_true',
                        help='Overwrite original file.')

    parser.add_argument('--critic-template', type=argparse.FileType('r'),
                        help='Custom template of CSS and JS for CriticMarkup in diff mode.')
    # parser.add_argument('--print-default-critic-template', type=argparse.FileType('w'), default=sys.stdout,
    #                     help='Custom template of CSS and JS for CriticMarkup in diff mode.')
    parser.add_argument('--engine', default='markdown',
                        help='If specified, convert markdown to HTML using the specified engine. Default: markdown. Valid: markdown, markdown2, panflute, pypandoc.')

    parser.add_argument('-m', '--critic-mode', default='diff',
                        help='Specify critic mode. Default: diff. Valid: a/accept, r/reject, d/diff, m/markup.')

    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    args = parser.parse_args()

    # from-format (args.input is a file descriptor)
    if not args.from_format:
        if args.input.name != '<stdin>':
            args.from_format = normalize_format(os.path.splitext(args.input.name)[1][1:])
        else:
            print("No input file extension nor from-format specified. Default to markdown.", file=sys.stderr)
            args.from_format = 'markdown'

    body = args.input.read()

    if args.inplace:
        if args.input.name == '<stdin>':
            print('Cannot perform inplace to stdin!', file=sys.stderr)
            exit(1)
        else:
            args.output = args.input.name

    args.input.close()
    

    # output-format (remember args.output is a string)
    try:
        output_format = normalize_format(os.path.splitext(args.output)[1][1:])
    except TypeError:
        print("No output file extension nor to-format specified. Default to HTML.", file=sys.stderr)
        output_format = 'html'

    if args.to is None:
        args.to = output_format

    is_binary = output_format in ("odt", "docx", "epub", "epub3", "pdf")

    if not is_binary:
        args.output = sys.stdout if args.output is None else open(args.output, 'w')
    elif args.output is None:
        print('Cannot output binary format to stdout', file=sys.stderr)
        exit(1)

    return args, body, output_format, is_binary


def cli():
    main(*get_args())



from __future__ import print_function

import re
import sys

from .template import CSS, NAV, JS

ADD_EDIT = re.compile(r'(?s)\{\+\+(.*?)\+\+[ \t]*(\[(.*?)\])?[ \t]*\}')
DEL_EDIT = re.compile(r'(?s)\{\-\-(.*?)\-\-[ \t]*(\[(.*?)\])?[ \t]*\}')
COMM_EDIT = re.compile(r'(?s)\{\>\>(.*?)\<\<[ \t]*(\[(.*?)\])?[ \t]*\}')
MARK_EDIT = re.compile(r'(?s)\{\=\=(.*?)\=\=[ \t]*(\[(.*?)\])?[ \t]*\}')
SUB_EDIT = re.compile(r'''(?s)\{\~\~(?P<original>(?:[^\~\>]|(?:\~(?!\>)))+)\~\>(?P<new>(?:[^\~\~]|(?:\~(?!\~\})))+)\~\~\}''')


def criticmarkup_accept_filter(body):
    body = ADD_EDIT.sub(r'\1', body)
    body = DEL_EDIT.sub(r'', body)
    body = SUB_EDIT.sub(r'\2', body)

    body = MARK_EDIT.sub(r'\1', body)
    return COMM_EDIT.sub(r'', body)


def criticmarkup_reject_filter(body):
    body = ADD_EDIT.sub(r'', body)
    body = DEL_EDIT.sub(r'\1', body)
    body = SUB_EDIT.sub(r'\1', body)

    body = MARK_EDIT.sub(r'\1', body)
    return COMM_EDIT.sub(r'', body)


def criticmarkup_tex_diff_filter(body):
    body = ADD_EDIT.sub(r'\\added{\1}', body)
    body = DEL_EDIT.sub(r'\\deleted{\1}', body)
    body = SUB_EDIT.sub(r'\\replaced{\2}{\1}', body)

    body = MARK_EDIT.sub(r'\\highlight{\1}', body)
    return COMM_EDIT.sub(r'\\comment{\1}', body)


def criticmarkup_html_diff_filter(body):

    def deletionProcess(group_object):
        replaceString = ''
        if group_object.group('value') == '\n\n':
            replaceString = "<del>&nbsp;</del>"
        else:
            replaceString = '<del>' + group_object.group('value').replace("\n\n", "&nbsp;") + '</del>'
        return replaceString

    def subsProcess(group_object):
        delString = '<del>' + group_object.group('original') + '</del>'
        insString = '<ins>' + group_object.group('new') + '</ins>'
        newString = delString + insString
        return newString

    # Converts Addition markup to HTML
    def additionProcess(group_object):
        replaceString = ''

        # Is there a new paragraph followed by new text
        if group_object.group('value').startswith('\n\n') and group_object.group('value') != "\n\n":
            replaceString = "\n\n<ins class='critic' break>&nbsp;</ins>\n\n"
            replaceString = replaceString + '<ins>' + group_object.group('value').replace("\n", " ")
            replaceString = replaceString + '</ins>'

        # Is the addition just a single new paragraph
        elif group_object.group('value') == "\n\n":
            replaceString = "\n\n<ins class='critic break'>&nbsp;" + '</ins>\n\n'

        # Is it added text followed by a new paragraph?
        elif group_object.group('value').endswith('\n\n') and group_object.group('value') != "\n\n":
            replaceString = '<ins>' + group_object.group('value').replace("\n", " ") + '</ins>'
            replaceString = replaceString + "\n\n<ins class='critic break'>&nbsp;</ins>\n\n"

        else:
            replaceString = '<ins>' + group_object.group('value').replace("\n", " ") + '</ins>'

        return replaceString

    def highlightProcess(group_object):
        replaceString = '<span class="critic comment">' + group_object.group('value').replace("\n", " ") + '</span>'
        return replaceString

    def markCommProcess(group_object):
        replaceString = '<mark>' + group_object.group('value') + '</mark><span class="critic comment">' + group_object.group('comment').replace("\n", " ") + '</span>'
        return replaceString

    def markProcess(group_object):
        replaceString = '<mark>' + group_object.group('value') + '</mark>'
        return replaceString

    add_pattern = r'''(?s)\{\+\+(?P<value>.*?)\+\+[ \t]*(\[(?P<meta>.*?)\])?[ \t]*\}'''

    del_pattern = r'''(?s)\{\-\-(?P<value>.*?)\-\-[ \t]*(\[(?P<meta>.*?)\])?[ \t]*\}'''

    comm_pattern = r'''(?s)\{\>\>(?P<value>.*?)\<\<\}'''

    subs_pattern = r'''(?s)\{\~\~(?P<original>(?:[^\~\>]|(?:\~(?!\>)))+)\~\>(?P<new>(?:[^\~\~]|(?:\~(?!\~\})))+)\~\~\}'''

    mark_comm_pattern = r'''(?s)\{\=\=(?P<value>.*?)\=\=\}\{\>\>(?P<comment>.*?)\<\<\}'''

    mark_pattern = r'''(?s)\{\=\=(?P<value>.*?)\=\=\}'''

    body = re.sub(del_pattern, deletionProcess, body, flags=re.DOTALL)

    body = re.sub(add_pattern, additionProcess, body, flags=re.DOTALL)

    body = re.sub(mark_comm_pattern, markCommProcess, body, flags=re.DOTALL)

    body = re.sub(mark_pattern, markProcess, body, flags=re.DOTALL)

    # comment processing must come after highlights
    body = re.sub(comm_pattern, highlightProcess, body, flags=re.DOTALL)

    return re.sub(subs_pattern, subsProcess, body, flags=re.DOTALL)


def markdown_filter(body, engine):

    def _markdown2_filter(body):
        from markdown2 import markdown
        return markdown(body, extras=['footnotes', 'fenced-code-blocks', 'cuddled-lists', 'code-friendly'])

    def _markdown_filter(body):
        from markdown import markdown
        return markdown(body, extensions=['extra', 'codehilite', 'meta'])

    engines = ('markdown2', 'markdown') if engine == 'markdown2' else ('markdown', 'markdown2')
    engine_function = {
        'markdown': _markdown_filter,
        'markdown2': _markdown2_filter
    }

    for i, engine in enumerate(engines):
        try:
            # i != 0 means failing last time
            if i != 0:
                print('Use {} instead.'.format(engine), file=sys.stderr)
            return engine_function[engine](body)
        except:
            print('Cannot use {}.'.format(engine), file=sys.stderr)

    print('Stop converting and output original format instead.', file=sys.stderr)

    return body


def pandoc_filter(body, input_format, output_format, standalone, engine, outputfile=None):
    extra_args = ['-s'] if standalone else []

    def panflute_filter(body, input_format, output_format, extra_args, outputfile):
        from panflute import convert_text
        return convert_text(body, input_format=input_format, output_format=output_format, extra_args=extra_args)

    def pypandoc_filter(body, input_format, output_format, extra_args, outputfile):
        from pypandoc import convert_text
        return convert_text(body, output_format, input_format, extra_args=extra_args, outputfile=outputfile)

    engines = ('panflute', 'pypandoc') if engine == 'panflute' else ('pypandoc', 'panflute')
    engine_function = {
        'panflute': panflute_filter,
        'pypandoc': pypandoc_filter
    }

    for i, engine in enumerate(engines):
        try:
            # i != 0 means failing last time
            if i != 0:
                print('Use {} instead.'.format(engine), file=sys.stderr)
            return engine_function[engine](body, input_format, output_format, extra_args, outputfile)
        except:
            print('Cannot use {}.'.format(engine), file=sys.stderr)

    print('Stop converting and output original format instead.', file=sys.stderr)

    return body


def html_filter(body, template, mode, standalone):

    def enclose(body, value, id=None):
        return ['<{} id="{}">'.format(value, id)] + body + ['</{}>'.format(value)] if id else ['<{}>'.format(value)] + body + ['</{}>'.format(value)]

    # convert to lists
    body = [body]

    if template:
        head = [template.read()]
    elif mode == 'm':
        head = [CSS, JS]
    elif mode == 'd':
        head = [CSS, NAV, JS]
    else:
        head = []

    body = (
        ['<!DOCTYPE html>'] + enclose(enclose(head, 'head') + enclose(enclose(body, 'div', id='wrapper'), 'body'), 'html')
        if head else
        ['<!DOCTYPE html>'] + enclose(enclose(body, 'body'), 'html')
    ) if standalone else (
        head + enclose(body, 'div', id='wrapper')
        if head else
        body
    )

    return '\n\n'.join(body)



CSS = '''<style>
	#wrapper {
		padding-top: 30px !important;
	}

	#criticnav {
		position: fixed;
		top: 0;
		left: 0;
		width: 100%;
		box-shadow: 0 1px 1px 1px #777;
		margin: 0;
		padding: 0;
		background-color: white;
		font-size: 12px;
	}

	#criticnav ul {
		list-style-type: none;
		width: 90%;
		margin: 0 auto;
		padding: 0;
	}

	#criticnav ul li {
		display: block;
		width: 33%;
		text-align: center;
		padding: 10px 0 5px!important;
		margin: 0 !important;
		line-height: 1em;
		float: left;
		border-left: 1px solid #ccc;
		text-transform: uppercase;
	}

	#criticnav ul li:before {
		content: none !important;
	}

	#criticnav ul li#edited-button {
		border-right: 1px solid #ccc;
	}

	#criticnav ul li.active {
		background-image: -webkit-linear-gradient(top, white, #cccccc)
	}

	.original del {
			text-decoration: none;
	}	

	.original ins,
	.original span.popover,
	.original ins.break {
		display: none;
	}

	.edited ins {
			text-decoration: none;
	}	

	.edited del,
	.edited span.popover,
	.edited ins.break {
		display: none;
	}

	.original mark,
	.edited mark {
		background-color: transparent;
	}

	.markup mark {
	    background-color: #fffd38;
	    text-decoration: none;
	}

	.markup del {
	    background-color: #f6a9a9;
	    text-decoration: none;
	}

	.markup ins {
	    background-color: #a9f6a9;
	    text-decoration: none;
	}

	.markup ins.break {
		display: block;
		line-height: 2px;
		padding: 0 !important;
		margin: 0 !important;
	}

	.markup ins.break span {
		line-height: 1.5em;
	}

	.markup .popover {
		background-color: #4444ff;
		color: #fff;
	}

	.markup .popover .critic.comment {
	    display: none;
	}

	.markup .popover:hover span.critic.comment {
	    display: block;
	    position: absolute;
	    width: 200px;
	    left: 30%;
	    font-size: 0.8em; 
	    color: #ccc;
	    background-color: #333;
	    z-index: 10;
	    padding: 0.5em 1em;
	    border-radius: 0.5em;
	}

</style>
'''

NAV = '''<div id="criticnav"><ul>
	<li id="markup-button">Markup</li>
	<li id="original-button">Original</li>
	<li id="edited-button">Edited</li>
</ul></div>
'''

JS = '''<script src="https://cdnjs.cloudflare.com/ajax/libs/jquery/3.3.1/jquery.slim.min.js"></script>

<script type="text/javascript">

	function critic() {

		$('#firstdiff').remove();
		$('#wrapper').addClass('markup');
		$('#markup-button').addClass('active');
		$('ins.break').unwrap();
		$('span.critic.comment').wrap('<span class="popover" />');
		$('span.critic.comment').before('&#8225;');

	}  

	function original() {
		$('#original-button').addClass('active');
		$('#edited-button').removeClass('active');
		$('#markup-button').removeClass('active');

		$('#wrapper').addClass('original');
		$('#wrapper').removeClass('edited');
		$('#wrapper').removeClass('markup');
	}

	function edited() {
		$('#original-button').removeClass('active');
		$('#edited-button').addClass('active');
		$('#markup-button').removeClass('active');

		$('#wrapper').removeClass('original');
		$('#wrapper').addClass('edited');
		$('#wrapper').removeClass('markup');
	} 

	function markup() {
		$('#original-button').removeClass('active');
		$('#edited-button').removeClass('active');
		$('#markup-button').addClass('active');

		$('#wrapper').removeClass('original');
		$('#wrapper').removeClass('edited');
		$('#wrapper').addClass('markup');
	}

	var o = document.getElementById("original-button");
	var e = document.getElementById("edited-button");
	var m = document.getElementById("markup-button");

	window.onload = critic;
	o.onclick = original;
	e.onclick = edited;
	m.onclick = markup;

</script>
'''
